module matrix_interface
  use types; use consts_dp
  use ew_parameters
  use shower_routines
  use sub_defs_io
  use rad_tools
  use hoppet_v1
  use expansion
  use masters
  use histogram_routines
  use iso_c_binding
  use pdfs_tools
  use seeds
  use matching

  implicit none

  public :: gen_rad_matrix, initialise_radish, book_matrix_histogram, print_one_histo
  public :: do_matching, do_matching_matrix
contains

  subroutine initialise_radish(nev, pdf_name, pdf_namelength, pdf_member, collider, colliderlenght, proc, proclength, rts, xQ, p, jet_radius, jet_veto, res_order, observable, obslength, seed) BIND(C)
    use iso_c_binding
    implicit none
    integer(c_int), value :: nev, pdf_member
    integer(c_int), value :: pdf_namelength, proclength, colliderlenght, obslength
    character(kind = c_char), intent(in) :: pdf_name, collider, proc, observable
    real(c_double), value  :: rts, xQ, p, jet_radius, jet_veto
    real(c_double) :: M, Q_in
    integer(c_int), value  :: res_order, seed

    ! check whether observable is correctly initialized
    if ( (observable(1:obslength) == 'ptB') .or. (observable(1:obslength) == 'ptj') .or. (observable(1:obslength) == 'phi') .or. (observable(1:obslength) == 'JV_ptB') .or. (observable(1:obslength) == 'JV_ptj')  ) then
      continue
    else
      call wae_error("unrecognized observable: "//trim(observable(1:obslength)))
    endif

    ! initialise LHAPDF
    call init_pdfs_from_LHAPDF(pdf_name(1:pdf_namelength), pdf_member)

    ! initialise radish parameters
    ! default values chosen for the invariant mass window (no cuts) and jet radius (not used) and default calls to turn off dynamical scales, dryrun, analytic expansion

    ! set seed
    call setseeds(seed,.false.)

    !set M to 100d0 for initialisation. Set Q to M*xQ such that xQ is correctly initialised in the code
    !mass, factorisation and renormalisation scales are dynamically allocated. They are set to be equal to M in initialisation
    M = 100d0
    Q_in = M*xQ

    call init_proc()

    !  dynamical,dryrun,analytic_expansion are set to false
    !  the cuts are set to arbitrarily large values
    ! TODO decide if setting dynamical scales to false is problematic
    call set_process_and_parameters(collider(1:colliderlenght), proc(1:proclength), rts, M, M, M, Q_in, p, res_order, 1d0, rts, 0d0, 99d0, jet_radius, jet_veto, "none", observable(1:obslength), .false., .false., .false., 'g', 'g' )

    ! initialise shower
    call init_shower(res_order,nev)

    ! initialise expansion:
    ! the Master Integrals are defined for the observable itself pt, or phi*
    call init_expansion()
  end subroutine initialise_radish


    ! call init_expansion()

    ! ! allocate histograms (e.g. 1 in the following example)
    ! i = 1
    ! call book_histogram(cs%observable, cs%observable, i, v)
    ! ! set the index that represents the scale (e.g. 1 for a single scale per run)
    ! call set_scale_index(i)
    ! call allocate_master_integrals(cs%observable, cs%observable, size(histo_resummed(i)%values%right(:)))


  subroutine book_matrix_histogram(name, namelength, obs, obslength, n_scales, bins_in, nbins) BIND(C)
    use iso_c_binding
    implicit none
    integer(c_int), value                      :: namelength, obslength, n_scales, nbins
    character(kind = c_char), intent(in)       :: name, obs
    real(c_double), dimension(*), intent(in), target :: bins_in
    real(c_double), pointer                          :: bins(:)
    integer(c_int)                             :: i

    bins => bins_in(1:nbins)
    do i = 1, n_scales
      call book_histogram(name(1:namelength), obs(1:obslength), i, bins)
      call allocate_master_integrals(name(1:namelength), obs(1:obslength), nbins)
    end do

  end subroutine book_matrix_histogram

  !print histogram histo_number to disk
  subroutine print_one_histo(histo_number, n_born, n_events) BIND(C)
    use iso_c_binding
    implicit none
    integer(c_int) :: histo_number
    character(5) :: histo_number_string
    character(100) :: histo_name, histo_folder
    character(200) :: histo_file, histo_file_lumi
    integer(c_int) :: bin_length
    integer(c_int) :: iunit, i, n_born, n_events
    real(c_double) :: left
    real(c_double) :: weight, error

    ! fetch name of the histogram and construct output name
    histo_name = histo_resummed(histo_number)%values%name
    if (histo_resummed(histo_number)%values%scale .ge. 10) then
        write(histo_number_string, "(I2)") histo_resummed(histo_number)%values%scale
    else 
        write(histo_number_string, "(I1)") histo_resummed(histo_number)%values%scale
    endif
    call system("mkdir -p RadISH_output")
    histo_folder = trim("RadISH_output/scale.")//trim(histo_number_string)
    call system("mkdir -p "//histo_folder)
    histo_file = trim(histo_folder)//"/"//trim(histo_name)//".dat"
    histo_file_lumi = trim(histo_folder)//"/"//trim(histo_name)//"--lum.dat"

    !get number of bins
    bin_length = size(histo_resummed(histo_number)%values%right)

    !now print
    iunit = 20
    open(iunit,file=histo_file)
    write(iunit, *) "# obs_low obs_up res res_err exp1 exp1_err exp2 exp2_err exp3 exp3_err"
    do i=1,bin_length
      if(i==1) then
        left = 0
      else
        left = histo_resummed(histo_number)%values%right(i-1)
      endif

      weight = histo_resummed(histo_number)%values%weight(i)/real(n_born)/real(n_events)
      error = sqrt(abs(histo_resummed(histo_number)%values%error(i)))/real(n_born)/real(n_events)

      write(iunit,*) left, histo_resummed(histo_number)%values%right(i), weight, &
        & error, &
        & histo_expanded(histo_number)%values%nlo(i)/real(n_born), 0d0, &
        & histo_expanded(histo_number)%values%nnlo(i)/real(n_born), 0d0, &
        & histo_expanded(histo_number)%values%n3lo(i)/real(n_born), 0d0

    end do
    close(iunit)

    !now print
    iunit = 21
    open(iunit,file=histo_file_lumi)
    write(iunit, *) "# dummy dummy exp0, exp0_err, lumi0, lumi0_err, lumi1, lumi1_err, lumi2, lumi2_err"
    write(iunit,*) 0d0, 0d0, histo_expanded(histo_number)%values%lo(1)/n_born, 0d0, histo_luminosity(histo_number)%values%asympt/n_born, 0d0, histo_luminosity(histo_number)%values%as/n_born, 0d0, histo_luminosity(histo_number)%values%as2/n_born, 0d0
    close(iunit)

  end subroutine print_one_histo


  !print histogram histo_number to disk
  subroutine print_all_histos(n_born) BIND(C)
    use iso_c_binding
    implicit none
    integer(c_int) :: n_histos, i, n_events
    integer(c_int), value :: n_born

    n_histos = size(histo_resummed)
    call get_nev(n_events)

    do i=1,n_histos
      call print_one_histo(i,n_born,n_events)
    end do

  end subroutine print_all_histos

  !shower for a single phase-space points
!#######################################################################
  subroutine gen_rad_matrix(pborn_in, pdg_ids, npart_in, msqB_in, msqV1_in, msqV2_in, nf_in, jac_in, shower_wgt, scale_index_in) BIND(C)
    use iso_c_binding
    integer, value :: nf_in, npart_in, scale_index_in
    real(dp), intent(inout) :: jac_in
    integer :: i, j
    !real(dp), intent(out) :: sigma_res(0:lenv-1), sigma_err(0:lenv-1), sigma_exp(0:lenv-1,0:3), lumi_asympt(0:3)
    real(dp):: msqB(-nf_int:nf_int,-nf_int:nf_int),msqV1(-nf_int:nf_int,-nf_int:nf_int)
    real(dp):: msqB_in(0:(2*nf_in+1)*(2*nf_in)-1), msqV1_in(0:(2*nf_in+1)*(2*nf_in)-1), msqV2_in(0:(2*nf_in+1)*(2*nf_in)-1)
    real(dp):: msqV2(-nf_int:nf_int,-nf_int:nf_int)
    real(dp):: pborn(14,4)
    real(dp):: pborn_in(0:npart_in*4-1)
    integer(c_int):: pdg_ids(0:npart_in)
    real(dp):: pswt
    real(dp), intent(out) :: shower_wgt
    logical :: debug
    pborn(:,:)=0._dp
    msqV1(:,:)=0._dp
    msqV2(:,:)=0._dp
    msqB(:,:)=0._dp
    debug = .false. ! pass via command line?


    if( nf_in .ne. nf_int ) then
    write(*,*) nf_in, nf_int, npart_in
    write(*,*) "Error: mismatch in the number of flavour: nf_in ", nf_in, "different from nf_int ", nf_int
    stop
    endif

    do i = 1,npart_in
    do j = 1,4
    ! assume canonical momenta as input and change to MCFM convention here
    if (i.eq.1 .or. i.eq.2) then
       pborn(i,j) = -pborn_in((i-1)*4+j-1)
    else
       pborn(i,j) = pborn_in((i-1)*4+j-1)
    endif
    enddo
    enddo

    if(debug) then
    do i = 1,npart_in
    print *, "particle ",i, "momentum = ",pborn(i,:)
    enddo
    endif

    do i=-nf_int,nf_int
    do j=-nf_int,nf_int
    msqB(i,j) = msqB_in((2*nf_in+1)*(i+nf_in)+j+nf_in)
    msqV1(i,j) = msqV1_in((2*nf_in+1)*(i+nf_in)+j+nf_in)
    msqV2(i,j) = msqV2_in((2*nf_in+1)*(i+nf_in)+j+nf_in)
    enddo
    enddo

    ! call to shower - test
    ! set the binning flag to true
    call do_binning();
    call set_scale_index(scale_index_in)

    if( (cs%observable == 'ptj') .and. (cs%numerical_jetveto .eqv. .false.) ) then
      call jetveto(pborn,msqB,msqV1,msqV2,jac_in,shower_wgt,14)
    else
      call make_shower(pborn,msqB,msqV1,msqV2,jac_in,shower_wgt,14)
    endif

  end subroutine gen_rad_matrix

  subroutine do_matching(scheme, nbins, fixed_order, resum_order, sigma, sigma_err, exp_lumi_in, exp_lumi_in_err, pt, sigmabar_1, sigmabar_1_err, sigmabar_2, sigmabar_2_err, sigmabar_3, sigmabar_3_err, resummed, resummed_err, res_expanded_1, res_expanded_1_err, res_expanded_2, res_expanded_2_err, res_expanded_3, res_expanded_3_err, matched, fixed_order_result, expansion_result, normalised, damping, observable_in, scale) BIND(C)
    use iso_c_binding
    implicit none
    character(kind = c_char, len=1), intent(in) :: scheme(*)
    integer(c_int), intent(in), value :: nbins, fixed_order, resum_order
    real(c_double), intent(in) :: sigma(0:3), sigma_err(0:3), exp_lumi_in_err(0:3)
    real(c_double) :: exp_lumi_in(0:3)
    real(c_double) :: lumi_asympt(0:3)
    real(c_double):: pt(1:nbins) 
    real(c_double):: sigmabar_1(1:nbins), sigmabar_1_err(1:nbins)
    real(c_double):: sigmabar_2(1:nbins), sigmabar_2_err(1:nbins)
    real(c_double):: sigmabar_3(1:nbins), sigmabar_3_err(1:nbins)
    real(c_double):: resummed(1:nbins), resummed_err(1:nbins)
    real(c_double):: res_expanded_1(1:nbins), res_expanded_1_err(1:nbins)
    real(c_double):: res_expanded_2(1:nbins), res_expanded_2_err(1:nbins)
    real(c_double):: res_expanded_3(1:nbins), res_expanded_3_err(1:nbins)
    real(c_double):: norm_resum
    real(c_double), intent(out) :: matched(1:nbins), fixed_order_result(1:nbins), expansion_result(1:nbins)
    real(dp) :: sigmabar(1:fixed_order,nbins), resum_expanded(0:fixed_order,nbins)
    real(c_double) :: dummy(1:nbins)
    logical(c_bool), intent(in), value :: normalised, damping
    character(kind = c_char, len=1), intent(in) :: observable_in(*)
    real(c_double), intent(in), value :: scale
    integer :: match_len, obs_len
    character(len=:), allocatable :: matching_scheme
    character(len=:), allocatable :: observable

    match_len=0
    do
        if (scheme(match_len+1) == C_NULL_CHAR) exit
        match_len = match_len + 1
    end do

    obs_len=0
    do
        if (observable_in(obs_len+1) == C_NULL_CHAR) exit
        obs_len = obs_len + 1
    end do
    allocate(character(len=obs_len) :: observable)
    observable = transfer(observable_in(1:obs_len), observable)


    allocate(character(len=match_len) :: matching_scheme)
    matching_scheme = transfer(scheme(1:match_len), matching_scheme)

    expansion_result(:) = exp_lumi_in(0) + res_expanded_1(:)
    resum_expanded(0,:) = exp_lumi_in(0)
    resum_expanded(1,:) = res_expanded_1
    sigmabar(1,:) = sigmabar_1(:)
    if(fixed_order > 1) then
        sigmabar(2,:) = sigmabar_2(:)
        resum_expanded(2,:) = res_expanded_2
        expansion_result(:) = expansion_result(:) + res_expanded_2(:)
    endif
    if(fixed_order > 2) then
        sigmabar(3,:) = sigmabar_3(:)
        resum_expanded(3,:) = res_expanded_3
        expansion_result(:) = expansion_result(:) + res_expanded_3(:)
    endif

    lumi_asympt(:) = zero
    lumi_asympt(0) = exp_lumi_in(1)
    lumi_asympt(1) = exp_lumi_in(2)
    lumi_asympt(2) = exp_lumi_in(3)

    norm_resum = exp_lumi_in(0) ! this sets the cross section units
    resummed = resummed * sigma(0)/norm_resum
    resum_expanded = resum_expanded * sigma(0)/norm_resum
    lumi_asympt = lumi_asympt * sigma(0)/norm_resum
    matched(:) = 0d0
    select case(matching_scheme)
        case("add")
            ! here dummy is an array of size(sigma);
            matched = matching_R(sigma, dummy, resummed, resum_expanded, sigmabar)
        case("mult")
            ! here dummy is an array of size(sigma); the last two arguments are dummy variables (TODO:remove them in the public release)
            if (damping) then
               matched = matching_mult_damping(sigma, pt, resummed, resum_expanded, sigmabar, lumi_asympt, 0.5d0, 3d0, observable, scale)
            else
               matched = matching_mult_a(sigma, dummy, resummed, resum_expanded, sigmabar, lumi_asympt, 0.5d0, 3d0)
            endif
        case default
           call wae_error("matching: unrecognised matching scheme "//trim(matching_scheme))
    end select
    if (matching_scheme == 'add') then
        matched = matched / sum(sigma(0:fixed_order))
        expansion_result = expansion_result / sum(sigma(0:fixed_order))
    elseif (matching_scheme == 'mult') then
        matched = matched / sum(sigma(0:fixed_order))
        expansion_result = expansion_result / sum(sigma(0:fixed_order))
    endif

    select case(matching_scheme)
    case("add")
        fixed_order_result = fixed_order_schemes("R", sigma, sigmabar(1:fixed_order,:))
    case("mult")
        fixed_order_result = fixed_order_schemes("multa", sigma, sigmabar(1:fixed_order,:))
    case default
       call wae_error("matching: unrecognised matching scheme "//trim(matching_scheme))
    end select

    if (normalised.eqv..false.) then
        matched = matched * sum(sigma(0:fixed_order))
        fixed_order_result = fixed_order_result * sum(sigma(0:fixed_order))
        expansion_result = expansion_result * sum(sigma(0:fixed_order))
    end if

    deallocate(matching_scheme)
    deallocate(observable)

  end subroutine do_matching


  subroutine do_matching_matrix(scheme, nbins, fixed_order, resum_order, sigmatot, sigmatot_err, exp_lumi_in, exp_lumi_in_err, sigma_1, sigma_1_err, sigma_2, sigma_2_err, sigma_3, sigma_3_err, resummed, resummed_err, res_expanded_1, res_expanded_1_err, res_expanded_2, res_expanded_2_err, res_expanded_3, res_expanded_3_err, matched, fixed_order_result, expansion_result, normalised) BIND(C)
    use iso_c_binding
    implicit none
    character(kind = c_char, len=1), intent(in) :: scheme(*)
    integer(c_int), intent(in), value :: nbins, fixed_order, resum_order
    real(c_double), intent(in) :: sigmatot(0:3), sigmatot_err(0:3), exp_lumi_in_err(0:3)
    real(c_double) :: exp_lumi_in(0:3)
    real(c_double) :: lumi_asympt(0:3)
    real(c_double):: sigma_1(1:nbins), sigma_1_err(1:nbins)
    real(c_double):: sigma_2(1:nbins), sigma_2_err(1:nbins)
    real(c_double):: sigma_3(1:nbins), sigma_3_err(1:nbins)
    real(c_double):: resummed(1:nbins), resummed_err(1:nbins)
    real(c_double):: res_expanded_1(1:nbins), res_expanded_1_err(1:nbins)
    real(c_double):: res_expanded_2(1:nbins), res_expanded_2_err(1:nbins)
    real(c_double):: res_expanded_3(1:nbins), res_expanded_3_err(1:nbins)
    real(c_double):: norm_resum
    real(c_double), intent(out) :: matched(1:nbins), fixed_order_result(1:nbins), expansion_result(1:nbins)
    real(dp) :: sigma(1:fixed_order,nbins), resum_expanded(0:fixed_order,nbins), sigmabar(1:fixed_order,nbins)
    real(c_double) :: dummy(1:nbins)
    logical(c_bool), intent(in), value :: normalised
    integer :: match_len
    character(len=:), allocatable :: matching_scheme

    match_len=0
    do
        if (scheme(match_len+1) == C_NULL_CHAR) exit
        match_len = match_len + 1
    end do

    allocate(character(len=match_len) :: matching_scheme)
    matching_scheme = transfer(scheme(1:match_len), matching_scheme)

    resum_expanded(0,:) = exp_lumi_in(0)
    resum_expanded(1,:) = res_expanded_1
    expansion_result(:) = exp_lumi_in(0) + res_expanded_1(:)
    sigma(1,:) = sigma_1(:)
    sigmabar(1,:) = sigma_1(:) - sigmatot(1)
    if(fixed_order > 1) then
        sigma(2,:) = sigma_2(:)
        sigmabar(2,:) = sigma_2(:) - sigmatot(2)
        resum_expanded(2,:) = res_expanded_2
        expansion_result(:) = expansion_result(:) + res_expanded_2(:)
    endif
    if(fixed_order > 2) then
        sigma(3,:) = sigma_3(:)
        sigmabar(3,:) = sigma_3(:) - sigmatot(3)
        resum_expanded(3,:) = res_expanded_3
        expansion_result(:) = expansion_result(:) + res_expanded_3(:)
    endif

    lumi_asympt(:) = zero
    lumi_asympt(0) = exp_lumi_in(1)
    lumi_asympt(1) = exp_lumi_in(2)
    lumi_asympt(2) = exp_lumi_in(3)

    norm_resum = exp_lumi_in(0) ! this sets the cross section units
    resummed = resummed * sigmatot(0)/norm_resum
    resum_expanded = resum_expanded * sigmatot(0)/norm_resum
    lumi_asympt = lumi_asympt * sigmatot(0)/norm_resum
    matched(:) = 0d0
    select case(matching_scheme)
        case("add")
            ! here dummy is an array of size(sigma);
            matched = matching_R_matrix(sigmatot(0), dummy, resummed, resum_expanded, sigma)
        case("mult")
            ! here dummy is an array of size(sigma); the last two arguments are dummy variables (TODO:remove them in the public release)
            matched = matching_mult_a_matrix(sigmatot(0), dummy, resummed, resum_expanded, sigma, lumi_asympt, 0.5d0, 3d0)
        case default
           call wae_error("matching: unrecognised matching scheme "//trim(matching_scheme))
    end select
    if (matching_scheme == 'add') then
        matched = matched / sum(sigmatot(0:fixed_order))
        expansion_result = expansion_result / sum(sigmatot(0:fixed_order))
    elseif (matching_scheme == 'mult') then
        matched = matched / sum(sigmatot(0:fixed_order))
        expansion_result = expansion_result / sum(sigmatot(0:fixed_order))
    endif

    select case(matching_scheme)
    case("add")
        fixed_order_result = fixed_order_schemes("R", sigmatot, sigmabar(1:fixed_order,:))
    case("mult")
        fixed_order_result = fixed_order_schemes("multa", sigmatot, sigmabar(1:fixed_order,:))
    case default
       call wae_error("matching: unrecognised matching scheme "//trim(matching_scheme))
    end select

    if (normalised.eqv..false.) then
        matched = matched * sum(sigmatot(0:fixed_order))
        fixed_order_result = fixed_order_result * sum(sigmatot(0:fixed_order))
        expansion_result = expansion_result * sum(sigmatot(0:fixed_order))
    end if

    deallocate(matching_scheme)

  end subroutine do_matching_matrix

end module matrix_interface
