module shower_routines
  use types; use consts_dp
  use rad_tools; use qcd
  use random_dists
  use pdfs_tools
  use pdf_expansions
  use sub_defs_io
  use warnings_and_errors
  use histogram_routines
  use interpolation_radish
  use momenta_and_recoil
  use observables
  use sudakov_radiators
  use expansion
  use emsn_tools
  implicit none

  private  :: gen_emissions
  integer, private, save  :: res_order, nev, number_of_iterations_simple_lumi
  public :: make_shower, gen_events, init_shower, reset_dynamical_shower, jetveto_resummation, get_nev, make_shower_inclusive, update_dynamical_parameters, jetveto, jetveto_inclusive

contains

  !-----------------------------------------------------------------------
  subroutine get_nev(nevents)
    integer, intent(inout) :: nevents
    nevents=nev
   end subroutine get_nev

  !-----------------------------------------------------------------------
  subroutine init_shower(order,nevents)
    integer, intent(in) :: order, nevents
    res_order=order
    nev=nevents
    !--- initialise emissions' momenta
    call init_scmom(q,zero)

    !--- initialise coefficient functions
    call init_C_matxs()
  end subroutine init_shower

  !-----------------------------------------------------------------------
  subroutine reset_dynamical_shower()
    !--- initialise coefficient functions
    if (res_order > order_NLL .or. ( (res_order == order_NLL) .and. cs%prime) ) call reset_dynamical_C_matxs()
  end subroutine reset_dynamical_shower

  subroutine update_dynamical_parameters(new_mur,new_muf,new_Msinglet) BIND(C)
    use iso_c_binding
    use rad_tools
    implicit none

    real(c_double) :: new_muR, new_muF, new_Msinglet

    !--- recalculate if using dynamical scale:
    call reset_dynamical_parameters(new_muR,new_muF, new_Msinglet)
    call reset_dynamical_shower()
    call reset_dynamical_expansion()

  end subroutine update_dynamical_parameters

  subroutine do_binning() BIND(C)
    use iso_c_binding
    implicit none
    logical :: bin ! this flag allows to fill the histograms
    common/bin/bin
    bin = .true.
  end subroutine do_binning

  ! The following subroutine computes the H1 factor as needed by the resummation
  ! and expansion (i.e. including the various muR, Q, lndlbar dependences)
  ! H(2) amounts to the two-loop virtual corrections,
  ! the constant term of the Sudakov (see jatRad.nb) and renormalisation
  ! of lower order virtual corrections
  ! two-loop virtual corrections from Becher-Neubert 1205.3806 (obtained with nf=5, N=3)
  subroutine get_virtual(msqB, msqV1, msqV2)
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int)
    real(dp), intent(inout) :: msqV1(-nf_int:nf_int,-nf_int:nf_int)
    real(dp), intent(inout) :: msqV2(-nf_int:nf_int,-nf_int:nf_int) ! two-loop virtual corrections

    ! Remember there's no lndlbar dependence for kt ordering
    msqV2 = msqV2 + ((A(1)*B(1)*half+A(1)*pi*beta0/three)*(-cs%ln_Q2_M2)**3   &
         & + 1/8._dp*A(1)**2*(-cs%ln_Q2_M2)**4 + (half*(-A(2)+B(1)**2) &
         & + pi*beta0*(B(1)-A(1)*(-cs%ln_Q2_muR2)))*cs%ln_Q2_M2**2 &
         & + (-B(2)-B(1)*(two*pi*beta0*(-cs%ln_Q2_muR2)))*(-cs%ln_Q2_M2))*msqB &
         & - B(1)*(-cs%ln_Q2_M2)*msqV1 - half*A(1)*cs%ln_Q2_M2**2*msqV1 &
         ! add renormalisation scale depedence
         & - as_pow*four*(-half*(one+as_pow)*pisq*beta0**2*cs%ln_muR2_M2**2-pisq*beta1*cs%ln_muR2_M2)*msqB &
         & + two*(one+as_pow)*pi*beta0*cs%ln_muR2_M2*msqV1 &
         ! the next term is the cross product of the two scale dependences in H1
         & + two*pi*beta0*cs%ln_muR2_M2*cs%ln_Q2_M2*(-half*A(1)*cs%ln_Q2_M2 + B(1))*msqB*as_pow
    ! add the additional term from our Sudakov
    if (.not.cs%analytic_expansion) msqV2 = msqV2 + 8._dp*A(1)*pi*beta0/3._dp*zeta3*msqB

    msqV1 = msqV1 + (-half*A(1)*cs%ln_Q2_M2+B(1))*cs%ln_Q2_M2*msqB &
         ! add renormalisation scale dependence
         & + two*as_pow*pi*beta0*cs%ln_muR2_M2*msqB
    return
  end subroutine get_virtual

  !-----------------------------------------------------------------------
  subroutine make_shower(pborn,msqB,msqV1,msqV2,vegas_wgt,shower_wgt,npart,index)
    logical :: bin ! this flag tells whether VEGAS is warming up
    common/bin/bin
    integer, intent(in) :: npart
    real(dp), intent(in) :: pborn(npart,4)
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int), vegas_wgt
    real(dp), intent(inout) :: msqV1(-nf_int:nf_int,-nf_int:nf_int), msqV2(-nf_int:nf_int,-nf_int:nf_int)
    real(dp), intent(out) :: shower_wgt ! this contains the shower weights averaged over the shower events
     !--------------------------------------------------
    real(dp) :: L
    real(dp) :: lambda_v1
    integer  :: i
    ! ******************************************************************************
    ! one can play with the following parameters to modify the generation
    integer,  parameter :: nbinstt=500 ! number of points at which the exact luminosity is evaluated
    real(dp), parameter :: ttmin=1d-3, ttmax=100._dp ! range in tt for luminosity evaluation
    ! ******************************************************************************
    real(dp) :: tt(nbinstt)
    integer  :: histo_index
    real(dp) :: lumi_nll(nbinstt)
    real(dp) :: lumi_nnll(nbinstt)
    real(dp) :: lumi_n3ll(nbinstt)
    real(dp) :: p0_lumi(nbinstt), p0_p0_lumi(nbinstt)
    real(dp) :: wgt_lumi(0:3)
    real(dp), allocatable :: wgt_exp(:,:)
    real(dp) :: xx(1:2)
    integer, optional:: index  ! speed: if index is known, it avoids the need to cycle over all the histograms (useful if number of histograms is very large)

    ! deal with parton luminosity
    xx(1)=-2._dp*pborn(1,4)/cs%rts
    xx(2)=-2._dp*pborn(2,4)/cs%rts

    ! compute virtual correction with all scale dependences
    call get_virtual(msqB,msqV1,msqV2)
    if (res_order < order_NLL) return
    ! here compute parton luminosities:
    ! define arrays for luminosity and its derivative
    ! Use a logarithmic binning to describe the low v1 region well
    !-----
    lumi_nll    = zero
    lumi_nnll   = zero
    lumi_n3ll   = zero
    !-----
    p0_lumi     = zero
    p0_p0_lumi  = zero
    !-----
    do i=1,nbinstt
       tt(i) = exp(log(ttmin)+(i-1)*(log(ttmax)-log(ttmin))/(nbinstt-1))
       L=Ltilde_scalar(tt(i),cs%p)
       lambda_v1 = cs%alphas_muR*beta0*L
       if (res_order == order_NLL) then
          lumi_nll(i)  = lumi_NLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1)
       elseif (res_order == order_NNLL) then
          if (bin) lumi_nll(i)  = lumi_NLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1)
          !-----
          lumi_nnll(i) = lumi_NNLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1,msqV2)
          !
          if (bin) p0_lumi(i)   = P0_lumi_NLL(cs%muF*exp(-L), lambda_v1, xx(1), xx(2),msqB)
       elseif (res_order == order_NNNLL) then
         if (bin) lumi_nll(i)   = lumi_NLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2), msqB,msqV1)
          !-----
         if (bin) lumi_nnll(i)  = lumi_NNLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1,msqV2)
          !-----
          lumi_n3ll(i)  = lumi_NNNLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1,msqV2)
          !
         if (bin) p0_lumi(i)    = P0_lumi_NLL(cs%muF*exp(-L), lambda_v1, xx(1), xx(2), msqB)
          !
         if (bin) p0_p0_lumi(i) = P0_P0_lumi_NLL(cs%muF*exp(-L), lambda_v1, xx(1), xx(2), msqB)
       end if
    end do

    ! initialise shower weight to return to VEGAS
    shower_wgt=zero

    if (present(index)) then
      call gen_events(lumi_nll,lumi_nnll,lumi_n3ll,p0_lumi,p0_p0_lumi,tt, nbinstt, vegas_wgt, shower_wgt, index)
    else
      call gen_events(lumi_nll,lumi_nnll,lumi_n3ll,p0_lumi,p0_p0_lumi,tt, nbinstt, vegas_wgt, shower_wgt)
    endif

    ! build the counterterm (expansion) for the matching and bin it
    if (bin) then
      if (present(index)) then
        allocate(wgt_exp(0:3,size(histo_resummed(index)%values%right(:))))
        call expanded_sigma(index, histo_resummed(index)%values%right(:), res_order, msqB, msqV1, msqV2, x1=xx(1), x2=xx(2), resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
        call fill_expansion(index,  vegas_wgt*wgt_exp)
        call fill_luminosity(index, vegas_wgt*wgt_lumi)
        deallocate(wgt_exp)
      else
        do histo_index = 1, number_of_histograms
          allocate(wgt_exp(0:3,size(histo_resummed(histo_index)%values%right(:))))
          call expanded_sigma(histo_index, histo_resummed(histo_index)%values%right(:), res_order, msqB, msqV1, msqV2, x1=xx(1), x2=xx(2), resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
          call fill_expansion(histo_index,  vegas_wgt*wgt_exp)
          call fill_luminosity(histo_index, vegas_wgt*wgt_lumi)
          deallocate(wgt_exp)
        end do
      endif
    end if


  end subroutine make_shower
  !-----------------------------------------------------------------------

  subroutine make_shower_inclusive(index)
    !--------------------------------------------------
    real(dp) :: L
    real(dp) :: lambda_v1
    integer  :: i
    ! ******************************************************************************
    ! one can play with the following parameters to modify the generation
    integer,  parameter :: nbinstt=500 ! number of points at which the exact luminosity is evaluated
    real(dp), parameter :: ttmin=1d-3, ttmax=100._dp ! range in tt for luminosity evaluation
    ! ******************************************************************************
    real(dp) :: tt(nbinstt)
    integer  :: histo_index
    real(dp) :: lumi_nll(nbinstt)
    real(dp) :: lumi_nnll(nbinstt)
    real(dp) :: lumi_n3ll(nbinstt)
    real(dp) :: p0_lumi(nbinstt), p0_p0_lumi(nbinstt)
    real(dp) :: wgt_lumi(0:3)
    real(dp), allocatable :: wgt_exp(:,:)
    integer, optional:: index  ! speed: if index is known, it avoids the need to cycle over all the histograms (useful if number of histograms is very large)

    if (res_order < order_NLL) return
    ! here compute parton luminosities:
    ! define arrays for luminosity and its derivative
    ! Use a logarithmic binning to describe the low v1 region well
    !-----
    lumi_nll    = zero
    !-----
    lumi_nnll   = zero
    !-----
    lumi_n3ll   = zero
    !-----
    p0_lumi     = zero
    p0_p0_lumi  = zero
    !-----
    do i=1,nbinstt
       tt(i) = exp(log(ttmin)+(i-1)*(log(ttmax)-log(ttmin))/(nbinstt-1))
       L=Ltilde_scalar(tt(i),cs%p)
       lambda_v1 = cs%alphas_muR*beta0*L
       if (res_order == order_NLL) then
          lumi_nll(i)  = inclusive_lumi_NLL_loc(cs%muF*exp(-L), lambda_v1)
       elseif (res_order == order_NNLL) then
          lumi_nll(i)  = inclusive_lumi_NLL_loc(cs%muF*exp(-L), lambda_v1)
          !-----
          lumi_nnll(i) = inclusive_lumi_NNLL_loc(cs%muF*exp(-L),lambda_v1)
          !
          p0_lumi(i)   = inclusive_P0_lumi_NLL(cs%muF*exp(-L), lambda_v1)
       elseif (res_order == order_NNNLL) then
          lumi_nll(i)   = inclusive_lumi_NLL_loc(cs%muF*exp(-L), lambda_v1)
          !-----
          lumi_nnll(i)  = inclusive_lumi_NNLL_loc(cs%muF*exp(-L),lambda_v1)
          !-----
          lumi_n3ll(i)  = inclusive_lumi_NNNLL_loc(cs%muF*exp(-L),lambda_v1)
          !
          p0_lumi(i)    = inclusive_P0_lumi_NLL(cs%muF*exp(-L), lambda_v1)
          !
          p0_p0_lumi(i) = inclusive_P0_P0_lumi_NLL(cs%muF*exp(-L), lambda_v1)
       end if
    end do

    call do_binning()

    if (present(index)) then
      call gen_events(lumi_nll,lumi_nnll,lumi_n3ll,p0_lumi,p0_p0_lumi,tt,nbinstt,1d0,index=index)
    else
      call gen_events(lumi_nll,lumi_nnll,lumi_n3ll,p0_lumi,p0_p0_lumi,tt,nbinstt,1d0)
    endif

    if (present(index)) then
      allocate(wgt_exp(0:3,size(histo_resummed(index)%values%right(:))))
      call expanded_sigma_inclusive(index, histo_resummed(index)%values%right(:), res_order, resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
      call fill_expansion(index,  wgt_exp)
      call fill_luminosity(index, wgt_lumi)
      deallocate(wgt_exp)
    else
      do histo_index = 1, number_of_histograms
        allocate(wgt_exp(0:3,size(histo_resummed(histo_index)%values%right(:))))
        call expanded_sigma_inclusive(histo_index, histo_resummed(histo_index)%values%right(:), res_order, resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
        call fill_expansion(histo_index,  wgt_exp)
        call fill_luminosity(histo_index, wgt_lumi)
        deallocate(wgt_exp)
      end do
    endif

  end subroutine make_shower_inclusive

  subroutine gen_events(lumi_nll, lumi_nnll, lumi_n3ll, p0_lumi, p0_p0_lumi, tt, nbinstt, vegas_wgt, shower_wgt, index)
    implicit none
    logical :: bin ! this flag tells whether VEGAS is warming up
    common/bin/bin
    !--------------------------------------------------
    integer :: nbinstt
    real(dp) :: lumi_nll(nbinstt), lumi_nnll(nbinstt), lumi_n3ll(nbinstt)
    real(dp) :: dlumi_nll(nbinstt), dlumi_nnll(nbinstt), dlumi_n3ll(nbinstt)
    real(dp) :: p0_lumi(nbinstt), p0_p0_lumi(nbinstt)
    real(dp) :: tt(nbinstt)
    real(dp) :: lumi_array(1:8,size(tt)), interp_lumi_array(1:8), err_lumi_array(1:8)
    real(dp), intent(in) :: vegas_wgt
    real(dp), intent(out), optional :: shower_wgt
    !--------------------------------------------------
    integer  :: iev
    real(dp) :: rp, L
    real(dp) :: jakob, jakob_sp1, jakob_sp2, lambda_v1
    integer  :: i, j
    real(dp) :: lumi_v1, dlumi_v1 ! luminosities and its derivative
    real(dp) :: p0_lumi_v1, p0_p0_lumi_v1 ! these are the P0xLNLL and P0xP0xLNLL in our eq 3.15
    real(dp) :: drp, rs, rt, npow, thetapow, genpow, v1, zeta_sp1, zeta_sp2
    real(dp) :: dy_error
    integer  :: icorr, histo_index
    real(dp) :: eta_diff
    integer, parameter :: number_of_weights = 11
    real(dp) :: wgt(number_of_weights), obs(number_of_histograms, number_of_weights)
    integer, optional:: index  ! speed: if index is known, it avoids the need to cycle over all the histograms (useful if number of histograms is very large)

    dlumi_nll   = zero
    dlumi_nnll  = zero
    dlumi_n3ll  = zero

    ! get derivative of parton luminosities
    call dlumi_Born(tt,lumi_nll,dlumi_nll)
    call dlumi_Born(tt,lumi_nnll,dlumi_nnll)
    call dlumi_Born(tt,lumi_n3ll,dlumi_n3ll)

    ! add the Jakobians -d tt(i)/ d Ltilde, Ltilde = 1/p*log(tt(i)**(-p)+1)
    ! NB: only if using numerical derivatives !!
    ! d tt(i)/dLtilde = - tt^(1+p) (1+tt^-p)
    dlumi_nll    = dlumi_nll   * tt**(one+cs%p)*(one+tt**(-cs%p))
    dlumi_nnll   = dlumi_nnll  * tt**(one+cs%p)*(one+tt**(-cs%p))
    dlumi_n3ll   = dlumi_n3ll  * tt**(one+cs%p)*(one+tt**(-cs%p))

    lumi_array(1,:) = lumi_nll(:)
    lumi_array(2,:) = lumi_nnll(:)
    lumi_array(3,:) = lumi_n3ll(:)
    lumi_array(4,:) = p0_lumi(:)
    lumi_array(5,:) = p0_p0_lumi(:)
    lumi_array(6,:) = dlumi_nll(:)
    lumi_array(7,:) = dlumi_nnll(:)
    lumi_array(8,:) = dlumi_n3ll(:)

    ! Loop over the shower events
    do iev=1,nev
      ! initialise observable values and weight
      obs = zero
      wgt = zero

      ! initalise Jacobians and luminosity prefactors
      jakob=zero
      jakob_sp1=zero
      jakob_sp2=zero

      lumi_v1=zero; dlumi_v1=zero ! These are luminosities evaluated at kt1


      !--- here generate emissions and compute observables
      call gen_emissions(nems,ktpow,jakob,v1,jakob_sp1,jakob_sp2,zeta_sp1,zeta_sp2,icorr,eta_diff)

      ! compute shower ingredients
      L=Ltilde_scalar(v1/cs%Q,cs%p)
      lambda_v1=cs%alphas_muR*L*beta0
      ! cutoff at the Landau pole
      if (lambda_v1 >= half) cycle

      drp = dRadp_v(L,res_order+1) ! include derivative of g3 for NNLL/ g4 at N3LL
      rs  = Rads_v(L,res_order+1)  ! include second derivative of g2 for NNLL
      rt  = Radt_v(L,res_order+1)  ! include third derivative of g1 for NNLL (in phi dependence)
      rp  = Radp_v(L)


      call interpolatemulti(tt(:), lumi_array, v1/cs%Q, interp_lumi_array, err_lumi_array, 3, size(interp_lumi_array))

      if (res_order == order_NLL) then
        !  call interpolate(tt(:), lumi_nll(:),   v1/cs%Q, lumi_v1, dy_error, 3)
        !  call interpolate(tt(:), dlumi_nll(:),  v1/cs%Q, dlumi_v1, dy_error, 3)
        lumi_v1 = interp_lumi_array(1)
        dlumi_v1 = interp_lumi_array(6)
      elseif (res_order == order_NNLL) then
        !  call interpolate(tt(:), lumi_nnll(:),   v1/cs%Q, lumi_v1, dy_error, 3)
        !  call interpolate(tt(:), dlumi_nnll(:),  v1/cs%Q, dlumi_v1, dy_error, 3)
        lumi_v1 = interp_lumi_array(2)
        dlumi_v1 = interp_lumi_array(7)
      elseif (res_order == order_NNNLL) then
        !  call interpolate(tt(:), lumi_n3ll(:),   v1/cs%Q, lumi_v1, dy_error, 3)
        !  call interpolate(tt(:), dlumi_n3ll(:),  v1/cs%Q, dlumi_v1, dy_error, 3)
        lumi_v1 = interp_lumi_array(3)
        dlumi_v1 = interp_lumi_array(8)
      end if

      wgt(1) = jakob * Sudakov(L,res_order)*((rp + drp) * lumi_v1 + dlumi_v1)

      if (bin) then
        ! cleanup the cached variables
        call reset_caching()

       !--- here compute the observable on the NLL ensemble
       if (present(index) ) then
        call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,1), special1=.false., special2=.false., inclusive=.false., res_order=order_NLL)
       else
        do histo_index = 1, number_of_histograms
          call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,1), special1=.false., special2=.false., inclusive=.false., res_order=order_NLL)
        end do
       endif

       p0_lumi_v1=zero; p0_p0_lumi_v1=zero;
       ! --- NNLL CORRECTIONS ---
       ! add the NNLL corrections:
       if (jakob.ne.zero.and.jakob_sp1.ne.zero) then
         if (res_order > order_NLL) then
           !--- interpolate luminosity and compute weight
           if (res_order == order_NNLL) then
            !  call interpolate(tt(:), lumi_nll(:), v1/cs%Q, lumi_v1, dy_error, 3)
            !  call interpolate(tt(:), dlumi_nll(:), v1/cs%Q, dlumi_v1, dy_error, 3)
             lumi_v1 = interp_lumi_array(1)
             dlumi_v1 = interp_lumi_array(6)
             p0_lumi_v1 = zero
             p0_p0_lumi_v1 = zero
             drp = dRadp_v(L,order_NNLL)
             rs  = Rads_v(L)
             rt  = zero
           elseif (res_order == order_NNNLL) then
            !  call interpolate(tt(:), lumi_nnll(:), v1/cs%Q, lumi_v1, dy_error, 3) ! also C1 here
            !  call interpolate(tt(:), dlumi_nnll(:), v1/cs%Q, dlumi_v1, dy_error, 3) ! also C1 here
            !  call interpolate(tt(:), p0_lumi(:), v1/cs%Q, p0_lumi_v1 , dy_error, 3)
            !  call interpolate(tt(:), p0_p0_lumi(:), v1/cs%Q, p0_p0_lumi_v1 , dy_error, 3)
             lumi_v1 = interp_lumi_array(2)
             dlumi_v1 = interp_lumi_array(7)
             p0_lumi_v1 = interp_lumi_array(4)
             p0_p0_lumi_v1 = interp_lumi_array(5)
             drp = dRadp_v(L,order_NNNLL)
             rs  = Rads_v(L,order_NNNLL) ! this includes also the g2 derivative now
             rt  = Radt_v(L,order_NNNLL)
           end if

           !----------------------------------------------------------------------------
           !----------------------------------------------------------------------------
           ! compute the weight for the soft-collinear and hard-collinear correction
           wgt(2) = wgt_sc_hc(L,lumi_v1,dlumi_v1,p0_lumi_v1,p0_p0_lumi_v1,rp,drp,rs,rt,zeta_sp1,res_order)
           wgt(2) = jakob_sp1 * wgt(2)
           wgt(3) = - wgt(2) ! this is for the counterterm

           if (present(index)) then
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,2), special1=.true., special2=.false., inclusive=.false., res_order=order_NLL) ! needs the observable in its NLL variant here
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,3), special1=.false., special2=.false., inclusive=.false., res_order=order_NLL)
           else
            do histo_index = 1, number_of_histograms
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,2), special1=.true., special2=.false., inclusive=.false., res_order=order_NLL) ! needs the observable in its NLL variant here
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,3), special1=.false., special2=.false., inclusive=.false., res_order=order_NLL)
            end do
           endif
           !----------------------------------------------------------------------------

           !----------------------------------------------------------------------------
           !----------------------------------------------------------------------------
           ! compute correlated corrections (to the CMW scheme)
           wgt(4) = wgt_corr(L,lambda_v1,lumi_v1,rp,rs,eta_diff,icorr) ! icorr is the emission of the ensemble correlated with the special one
           wgt(4) = jakob_sp1 * wgt(4)
           wgt(5) = - wgt(4)

           ! Add extra factor of two if the correlated emission is the first one
           if (icorr .eq. 1) then
              wgt(4) = two*wgt(4)
              wgt(5) = two*wgt(5)
           end if

           if (present(index)) then
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,4), special1=.true., special2=.false., inclusive=.false., res_order=res_order, &
            & icorr=icorr, eta_diff=eta_diff)
            !--- here compute observable with the special emission in the inclusive approximation
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,5), special1=.true., special2=.false., inclusive=.true., res_order=res_order, &
            & icorr=icorr, eta_diff=eta_diff)
           else
            do histo_index = 1, number_of_histograms
              !--- here compute observable with the special emission
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,4), special1=.true., special2=.false., inclusive=.false., res_order=res_order, &
              & icorr=icorr, eta_diff=eta_diff)
              !--- here compute observable with the special emission in the inclusive approximation
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,5), special1=.true., special2=.false., inclusive=.true., res_order=res_order, &
              & icorr=icorr, eta_diff=eta_diff)
            end do
           endif
           !----------------------------------------------------------------------------
           !----------------------------------------------------------------------------

           !----------------------------------------------------------------------------
           !----------------------------------------------------------------------------
           ! compute clustering corrections
           wgt(6) = wgt_clus(L,lambda_v1,lumi_v1,rp,rs,eta_diff) ! icorr is the emission of the ensemble correlated with the special one
           wgt(6) = jakob_sp1 * wgt(6)
           wgt(7) = - wgt(6)

           ! Add extra factor of two if the correlated emission is the first one
           if (icorr .eq. 1) then
              wgt(6) = two*wgt(6)
              wgt(7) = two*wgt(7)
           end if

           if (present(index)) then
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,6), special1=.true., special2=.false., inclusive=.false., res_order=res_order, &
            & icorr=icorr, eta_diff=eta_diff)
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,7), special1=.true., special2=.false., inclusive=.false., res_order=order_NLL)
           else
            do histo_index = 1, number_of_histograms
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,6), special1=.true., special2=.false., inclusive=.false., res_order=res_order, &
              & icorr=icorr, eta_diff=eta_diff)
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,7), special1=.true., special2=.false., inclusive=.false., res_order=order_NLL)
            end do
           endif
           !----------------------------------------------------------------------------
           !----------------------------------------------------------------------------
         end if
       end if


       ! --- N3LL CORRECTIONS ---
       if (jakob.ne.zero.and.jakob_sp2.ne.zero) then
         if (res_order > order_NNLL) then
           !--- interpolate NLL luminosity and compute weight
           if (res_order == order_NNNLL) then
            !  call interpolate(tt(:), lumi_nll(:), v1/cs%Q, lumi_v1, dy_error, 3)
            !  call interpolate(tt(:), dlumi_nll(:), v1/cs%Q, dlumi_v1, dy_error, 3)
             lumi_v1 = interp_lumi_array(1)
             dlumi_v1 = interp_lumi_array(6)
             drp = dRadp_v(L,order_NNLL)
             rs  = Rads_v(L)
           end if

           !----------------------------------------------------------------------------
           !----------------------------------------------------------------------------
           ! compute the weight for the soft-collinear and hard-collinear correction
           wgt(8)  = wgt_sc_hc_2emsn(L, lumi_v1, dlumi_v1, p0_p0_lumi_v1, rp, drp, rs, zeta_sp1, zeta_sp2)
           wgt(8)  = jakob_sp2 * wgt(8)
           wgt(9)  = - wgt(8)
           wgt(10) = - wgt(8)
           wgt(11) = wgt(8)

           !--- compute observable with the special emissions
           if (present(index)) then
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,8), special1=.true., special2=.true., inclusive=.false., res_order=order_NNNLL)
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,9), special1=.true., special2=.false., inclusive=.false., res_order=order_NNNLL)
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,10), special1=.false., special2=.true., inclusive=.false., res_order=order_NNNLL)
            call compute_observable(trim(histo_resummed(index)%values%obs),obs(index,11), special1=.false., special2=.false., inclusive=.false., res_order=order_NNNLL)
           else
            do histo_index = 1, number_of_histograms
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,8), special1=.true., special2=.true., inclusive=.false., res_order=order_NNNLL)
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,9), special1=.true., special2=.false., inclusive=.false., res_order=order_NNNLL)
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,10), special1=.false., special2=.true., inclusive=.false., res_order=order_NNNLL)
              call compute_observable(trim(histo_resummed(histo_index)%values%obs),obs(histo_index,11), special1=.false., special2=.false., inclusive=.false., res_order=order_NNNLL)
            end do
          endif

         end if
         !----------------------------------------------------------------------------
         !----------------------------------------------------------------------------
       end if

       ! now fill histograms
       if (present(index)) then
        call fill_histogram(index, obs(index,1), wgt(1)*vegas_wgt)
        if(res_order == order_NNLL) then
           call fill_histogram_array_smooth(index, obs(index,2:7), wgt(2:7)*vegas_wgt)
        else if(res_order == order_NNNLL) then
           call fill_histogram_array_smooth(index, obs(index,2:), wgt(2:)*vegas_wgt)
        endif
       else
        do histo_index = 1, number_of_histograms
          call fill_histogram(histo_index, obs(histo_index,1), wgt(1)*vegas_wgt)
          if(res_order == order_NNLL) then
             call fill_histogram_array_smooth(histo_index, obs(histo_index,2:7), wgt(2:7)*vegas_wgt)
          else if(res_order == order_NNNLL) then
             call fill_histogram_array_smooth(histo_index, obs(histo_index,2:), wgt(2:)*vegas_wgt)
          endif
        end do
       endif
      end if

      ! update the shower weight to be returned to VEGAS
      ! Pass the 1st emission contribution since the secondary corrections integrate
      ! to zero as v->inf.
      if( present(shower_wgt) ) then
        shower_wgt = shower_wgt + wgt(1)/nev ! this essentially corresponds to integrating inclusively over the radiation
      endif
   end do

  end subroutine gen_events


  ! Routine to generate a variable number of emissione (nems)
  ! with a cut on lnkti (lnv_max).
  subroutine gen_emissions(nems_out,ktpow_in,jakob,v1,jakob_sp1,jakob_sp2,zeta_sp1,zeta_sp2,icorr,eta_diff)
    integer, intent(out)     :: nems_out, icorr
    real(dp), intent(out) :: jakob, jakob_sp1, jakob_sp2, zeta_sp1, zeta_sp2, v1, eta_diff
    real(dp), intent(in)  :: ktpow_in ! at this stage we work in the soft-collinear limit
    real(dp), parameter   :: max_lnzeta=20._dp
    real(dp)              :: max_rp
    ! -------------------------------
    real(dp) :: u, ln_zeta, r, genpow, L, ransp_1, ransp_2
    integer :: i
    real(dp) :: rp, z
    real(dp) :: npow !>> this is used in the sampling of the special emissions (NNLL & N3LL)

    ! set max_rp for efficiency
    max_rp = 8._dp
    !>> can set to lower values for observables with kinematic cancellations
    if (cs%observable == 'ptB' .or. cs%observable == 'phi') max_rp = 5._dp

    nems_out=1
    ! generate the first emission
    r=ran()
    if (r.eq.zero.or.r.eq.one) return ! no solution in this case
    ! This assumes the use of modified logarithms in v1
    genpow=one
    v1=(r/(one-r))**genpow*cs%Q ! Q is a mass dimension scale here

    call gen_scmom(v1,ktpow_in,q(1))

    ! get jacobian factor for the first emission
    jakob = genpow/r/(one-r)

    ! add jakobian associated with modified logarithms in v1
    jakob = jakob*(cs%Q/v1)**cs%p/((cs%Q/v1)**cs%p + one)

    ! define L, R' using modified logs of degree cs%p
    L=Ltilde_scalar(v1/cs%Q,cs%p)
    rp  = Radp_v(L)

    ! veto events above max_rp
    ! This is an effective non-perturbative (NP) cutoff:
    ! a variation of max_rp estimates the size of NP corrections
    if (rp>max_rp) then
       !write(*,*) 'rp > max_rp', rp, max_rp, v1
       jakob=zero
       jakob_sp1=zero
       zeta_sp1=one
       jakob_sp2=zero
       zeta_sp2=one
       v1=1e-6
       eta_diff=one
       icorr=1
       return
    end if

    ! generate subsequent emissions
    ln_zeta = zero ! zeta=vi/v1
    ! prob \sim (v_{i+1}/v_i)^rp = u (random)
    ! lnv_{i+1} = lnv_{i} + ln(-u^{1/rp} = lnv + rd_exp(u) (minus due to the fact that lnv is positive)
    do
       ! generate secondary emissions according to the NLL probability
       ! exit if rp=0 (i.e. kt1->inf) --> no secondary emissions in this case
       if (rp==zero) exit
       u=rd_exp(rp)
       ln_zeta=ln_zeta + u
       if (ln_zeta > max_lnzeta) exit
       nems_out=nems_out+1

       ! generate soft-collinear emission
       if (nems_out == maxems) then
          write(*,*) "Warning: max number of emissions reached"
          exit ! This is dangerous, make sure it doesn't happen often
       end if
       call gen_scmom(exp(-ln_zeta)*v1,ktpow_in,q(nems_out))
    end do

    ! now generate special emissions: set the sampling power
    npow = 5._dp

    ! ---- FIRST SPECIAL EMISSION ----
    ! generate the first special emission (nems+1-th entry of q)
    ! parameter used to generate the special emission
    ransp_1 = ran()
    zeta_sp1= ransp_1**npow
    if (zeta_sp1.eq.zero) then
       zeta_sp1=one
       jakob_sp1=zero
    else
       jakob_sp1=jakob*npow/ransp_1
    end if
    call gen_scmom(zeta_sp1*v1,ktpow_in,q(nems_out+1))

    ! --- THIS BLOCK IS RELEVANT FOR CORRELATED/CLUSTERING NNLL CORRECTIONS ---
    ! The following stuff is needed for correlated/clustering corrections
    ! extract randomly which emission will be correlated with the special one
    icorr = ceiling(nems_out*ran())
    ! generate the rapidity difference between this emission and the special one
    z=ran()
    if (z==zero) then
       eta_diff=one
       jakob_sp1=zero
    else
       eta_diff = -log(z)
       if (ran()>half) eta_diff=-eta_diff
    end if
    ! --- END OF BLOCK RELEVANT FOR CORRELATED/CLUSTERING NNLL CORRECTIONS ---

    ! ---- SECOND SPECIAL EMISSION ----
    ! generate the second special emission (nems+2-th entry of q)
    ! parameter used to generate the special emission
    ransp_2 = ran()
    zeta_sp2= ransp_2**npow
    if (zeta_sp2.eq.zero) then
       zeta_sp2=one
       jakob_sp2=zero
    else
       jakob_sp2=jakob_sp1*npow/ransp_2
    end if
    call gen_scmom(zeta_sp2*v1,ktpow_in,q(nems_out+2))
    return
  end subroutine gen_emissions

  !-----------------------------------------------------------------------
  !-----------------------------------------------------------------------
  ! Compute NNLL and NNNLL weights:

  ! soft-collinear and hard-collinear
  function wgt_sc_hc(L,lumi_v1,dlumi_v1,p0_lumi_v1,p0_p0_lumi_v1,rp,drp,rs,rt,zeta_sp,res_order_in) result(wgt)
    real(dp), intent(in) :: L, lumi_v1, dlumi_v1, p0_lumi_v1, p0_p0_lumi_v1, rp, drp, rs, rt, zeta_sp
    integer, intent(in) :: res_order_in
    real(dp) :: wgt, phi_sp, phi1
    real(dp), parameter :: zeta_cut=1d-6

    if (zeta_sp<zeta_cut) then
          wgt = zero
    else
      if (res_order_in == order_NNLL) then
       wgt = Sudakov(L,order_NLL)*rp &
       & *(lumi_v1 * (drp + rs*log(one/zeta_sp)) + dlumi_v1)
     elseif (res_order_in == order_NNNLL) then
       wgt = Sudakov(L,order_NNLL)*( &
        & ((rp+drp)*lumi_v1 + dlumi_v1) * (drp + rs*log(one/zeta_sp) &
          & + half*rt*log(one/zeta_sp)**2) &
        & + (rp+drp)*(dlumi_v1 + p0_lumi_v1*log(one/zeta_sp)) &
        & + p0_p0_lumi_v1)
     end if
   end if
   return
  end function wgt_sc_hc

  ! soft-collinear and hard-collinear for two emissions (N3LL)
  function wgt_sc_hc_2emsn(L,lumi_v1,dlumi_v1,p0_p0_lumi_v1,rp,drp,rs,zeta_sp1,zeta_sp2) result(wgt)
    real(dp), intent(in) :: L, lumi_v1, dlumi_v1, p0_p0_lumi_v1, rp, drp,rs, zeta_sp1,zeta_sp2
    real(dp) :: wgt, phi_sp1,phi_sp2, phi1
    real(dp), parameter :: zeta_cut=1d-6

    if (zeta_sp1<zeta_cut.or.zeta_sp2<zeta_cut) then
          wgt = zero
    else
      wgt = half*Sudakov(L,order_NLL)*rp &
       & *(lumi_v1 * (drp + rs*log(one/zeta_sp1)) &
       & *(drp + rs*log(one/zeta_sp2)) &
       & + dlumi_v1*(rs*(log(one/zeta_sp1) + log(one/zeta_sp2)) + two*drp) &
       & + p0_p0_lumi_v1)
    end if
    return
  end function wgt_sc_hc_2emsn

  ! correlated (double-soft current)
  function wgt_corr(L,lambda_v1,lumi_v1,rp,rs,eta_diff,icorr) result(wgt)
    real(dp), intent(in) :: L, lambda_v1, lumi_v1, rp, rs, eta_diff
    integer,  intent(in) :: icorr
    real(dp) :: wgt, phi1
    real(dp) :: pjet(4,4)

    ! work in the c.o.m. of the collision
    pjet(:,1) = cs%Q*(/zero,zero,one,one/)
    pjet(:,2) = cs%Q*(/zero,zero,-one,one/)
    pjet(:,3) = (/q(icorr)%kt(1),q(icorr)%kt(2),zero,q(icorr)%ktmod/) ! placed along eta=0
    pjet(:,4) = (/q(nems+1)%kt(1),q(nems+1)%kt(2),&
         &q(nems+1)%ktmod*sinh(eta_diff),q(nems+1)%ktmod*cosh(eta_diff)/)

    ! splitting into two gluons
    wgt = half*ca*Mgg(pjet(:,1),pjet(:,2),pjet(:,3),pjet(:,4)) ! half is a symmetry factor for identical particles
    ! splitting into a qqbar pair
    wgt = wgt + half*nf*Mqq(pjet(:,1),pjet(:,2),pjet(:,3),pjet(:,4)) ! half is Tf=1/2

    wgt = wgt * nems*exp(abs(eta_diff))*Sudakov(L,order_NLL)*lumi_v1 &
         & *(lambda_v1/(pi*beta0)*rs)
    return
  end function wgt_corr

  ! clustering (abelian)
  function wgt_clus(L,lambda_v1,lumi_v1,rp,rs,eta_diff) result(wgt)
    real(dp), intent(in) :: L, lambda_v1, lumi_v1, rp, rs, eta_diff
    real(dp) :: wgt, phi1

    wgt = nems*exp(abs(eta_diff))*Sudakov(L,order_NLL)*lumi_v1 &
         & *(lambda_v1*A(1)/(pi*beta0)*rs)
    return
  end function wgt_clus

  !-----------------------------------------------------------------------
  ! the following routine implements the analytic resummation for the leading-jet pT
  subroutine jetveto(pborn,msqB,msqV1,msqV2,vegas_wgt,shower_wgt,npart,index)
    logical :: bin ! this flag tells whether VEGAS is warming up
    common/bin/bin
    integer, intent(in) :: npart
    real(dp), intent(in) :: pborn(npart,4)
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int), vegas_wgt
    real(dp), intent(inout) :: msqV1(-nf_int:nf_int,-nf_int:nf_int), msqV2(-nf_int:nf_int,-nf_int:nf_int)
    real(dp), intent(out) :: shower_wgt ! this contains the shower weights averaged over the shower events
    !--------------------------------------------------
    real(dp) :: L
    real(dp) :: lambda_v1
    integer  :: i
    ! ******************************************************************************
    ! one can play with the following parameters to modify the generation
    integer,  parameter :: nbinstt=500 ! number of points at which the exact luminosity is evaluated
    real(dp), parameter :: ttmin=1d-3, ttmax=100._dp ! range in tt for luminosity evaluation
    ! ******************************************************************************
    real(dp) :: tt(nbinstt), dy_error, histep, error
    real(dp) :: lumi_nll(nbinstt)
    real(dp) :: lumi_nnll(nbinstt)
    real(dp) :: xx(1:2)
    real(dp) :: wgt_lumi(0:3)
    real(dp), allocatable :: wgt_exp(:,:)
    integer:: histo_index
    integer, optional:: index  ! speed: if index is known, it avoids the need to cycle over all the histograms (useful if number of histograms is very large)

    ! deal with parton luminosity
    xx(1)=-2._dp*pborn(1,4)/cs%rts
    xx(2)=-2._dp*pborn(2,4)/cs%rts

    ! compute virtual correction with all scale dependences
    call get_virtual(msqB,msqV1,msqV2)
    if (res_order < order_NLL) return
    ! here compute parton luminosities:
    ! define arrays for luminosity and its derivative
    ! Use a logarithmic binning to describe the low v1 region well
    !-----
    lumi_nll    = zero
    !-----
    lumi_nnll   = zero
    !-----
    do i=1,nbinstt
      tt(i) = exp(log(ttmin)+(i-1)*(log(ttmax)-log(ttmin))/(nbinstt-1))
      L=Ltilde_scalar(tt(i),cs%p)
      lambda_v1 = cs%alphas_muR*beta0*L
      if (res_order == order_NLL) then
        lumi_nll(i)  = lumi_NLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1)
      elseif (res_order == order_NNLL) then
        lumi_nll(i)  = lumi_NLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1)
        !-----
        lumi_nnll(i)  = lumi_NNLL_Born(cs%muF*exp(-L),lambda_v1,xx(1),xx(2),msqB,msqV1,msqV2)
      end if
    end do

    ! initialise shower weight to return to VEGAS
    shower_wgt=zero

    if (present(index)) then
      call jetveto_resummation(lumi_nll, lumi_nnll, tt, nbinstt, vegas_wgt, shower_wgt, index=index)
    else
      call jetveto_resummation(lumi_nll, lumi_nnll, tt, nbinstt, vegas_wgt, shower_wgt)
    endif

    ! build the counterterm (expansion) for the matching and bin it
    if (bin) then
      if (present(index)) then
        allocate(wgt_exp(0:3,size(histo_resummed(index)%values%right(:))))
        call expanded_sigma_ptj(histo_resummed(index)%values%right(:), res_order, msqB, msqV1, msqV2, x1=xx(1), x2=xx(2), resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
        call fill_expansion(index, vegas_wgt*wgt_exp)
        call fill_luminosity(index, vegas_wgt*wgt_lumi)
        deallocate(wgt_exp)
      else
        do histo_index=1,number_of_histograms
          allocate(wgt_exp(0:3,size(histo_resummed(histo_index)%values%right(:))))
          call expanded_sigma_ptj(histo_resummed(histo_index)%values%right(:), res_order, msqB, msqV1, msqV2, x1=xx(1), x2=xx(2), resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
          call fill_expansion(histo_index, vegas_wgt*wgt_exp)
          call fill_luminosity(histo_index, vegas_wgt*wgt_lumi)
          deallocate(wgt_exp)
        end do
      endif
    end if

  end subroutine jetveto
  !-----------------------------------------------------------------------

  !-----------------------------------------------------------------------
  ! the following routine implements the analytic resummation for the leading-jet pT in the inclusive case
  subroutine jetveto_inclusive(index)
    !--------------------------------------------------
    real(dp) :: L
    real(dp) :: lambda_v1
    integer  :: i
    ! ******************************************************************************
    ! one can play with the following parameters to modify the generation
    integer,  parameter :: nbinstt=500 ! number of points at which the exact luminosity is evaluated
    real(dp), parameter :: ttmin=1d-3, ttmax=100._dp ! range in tt for luminosity evaluation
    ! ******************************************************************************
    real(dp) :: tt(nbinstt), dy_error
    real(dp) :: lumi_nll(nbinstt)
    real(dp) :: lumi_nnll(nbinstt)
    real(dp) :: xx(1:2)
    real(dp) :: wgt_lumi(0:3)
    real(dp), allocatable :: wgt_exp(:,:)
    integer:: histo_index
    integer, optional:: index  ! speed: if index is known, it avoids the need to cycle over all the histograms (useful if number of histograms is very large)

    if (res_order < order_NLL) return
    ! here compute parton luminosities:
    ! define arrays for luminosity and its derivative
    ! Use a logarithmic binning to describe the low v1 region well
    !-----
    lumi_nll    = zero
    !-----
    lumi_nnll   = zero
    !-----
    do i=1,nbinstt
      tt(i) = exp(log(ttmin)+(i-1)*(log(ttmax)-log(ttmin))/(nbinstt-1))
      L=Ltilde_scalar(tt(i),cs%p)
      lambda_v1 = cs%alphas_muR*beta0*L
      if (res_order == order_NLL) then
        lumi_nll(i)  = inclusive_lumi_NLL_loc(cs%muF*exp(-L),lambda_v1)
      elseif (res_order == order_NNLL) then
        lumi_nll(i)  = inclusive_lumi_NLL_loc(cs%muF*exp(-L),lambda_v1)
        !-----
        lumi_nnll(i)  = inclusive_lumi_NNLL_loc(cs%muF*exp(-L),lambda_v1)
      end if
    end do

    call do_binning()
    if (present(index)) then
      call jetveto_resummation(lumi_nll, lumi_nnll, tt, nbinstt, 1d0, index=index)
    else
      call jetveto_resummation(lumi_nll, lumi_nnll, tt, nbinstt, 1d0)
    endif

    ! build the counterterm (expansion) for the matching and bin it
    if (present(index) ) then
      allocate(wgt_exp(0:3,size(histo_resummed(index)%values%right(:))))
      call expanded_sigma_ptj_inclusive(histo_resummed(index)%values%right(:), res_order, resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
      call fill_expansion(index, wgt_exp)
      call fill_luminosity(index, wgt_lumi)
      deallocate(wgt_exp)
    else
      do histo_index = 1, number_of_histograms
        allocate(wgt_exp(0:3,size(histo_resummed(histo_index)%values%right(:))))
        call expanded_sigma_ptj_inclusive(histo_resummed(histo_index)%values%right(:), res_order, resummed_expanded=wgt_exp, lumi_asympt=wgt_lumi)
        call fill_expansion(histo_index, wgt_exp)
        call fill_luminosity(histo_index, wgt_lumi)
        deallocate(wgt_exp)
      end do
    endif
  end subroutine jetveto_inclusive

  subroutine jetveto_resummation(lumi_nll, lumi_nnll, tt, nbinstt, vegas_wgt, shower_wgt, index)
    logical :: bin ! this flag tells whether VEGAS is warming up
    common/bin/bin
    !--------------------------------------------------
    real(dp), intent(in) :: vegas_wgt
    real(dp), intent(out), optional :: shower_wgt ! this contains the shower weights averaged over the shower events
    !--------------------------------------------------
    integer  :: iev
    real(dp) :: wgt, rp, L
    real(dp) :: lambda_v1
    integer  :: i
    real(dp) :: lumi_v1
    ! ******************************************************************************
    integer :: nbinstt
    real(dp) :: tt(nbinstt), dy_error
    real(dp) :: lumi_nll(nbinstt)
    real(dp) :: lumi_nnll(nbinstt)
    integer:: histo_index
    integer, optional:: index  ! speed: if index is known, it avoids the need to cycle over all the histograms (useful if number of histograms is very large)

    if (present(index)) then
      lumi_v1=zero ! These are luminosities evaluated at kt1 !!
      do i=1,size(histo_resummed(index)%values%right(:))
        ! compute shower ingredients
        L=Ltilde_scalar(histo_resummed(index)%values%right(i)/cs%Q,cs%p)
        rp  = Radp_v(L)
        lambda_v1=cs%alphas_muR*L*beta0

        if (res_order == order_NLL) then
          call interpolate(tt(:), lumi_nll(:),   histo_resummed(index)%values%right(i)/cs%Q, lumi_v1, dy_error, 3)
          wgt = Sudakov(L,res_order)*lumi_v1
        else if (res_order == order_NNLL) then
          call interpolate(tt(:), lumi_nnll(:),   histo_resummed(index)%values%right(i)/cs%Q, lumi_v1, dy_error, 3)
          ! the following weight is from the original JetVHeto implementation
          ! write(*,*) 'jetveto_resummation: using original JetVHeto implementation'
          wgt = Sudakov(L,res_order)*lumi_v1*(1 + &
          & non_incl(cs%jet_radius,'all') * rp*two*cs%as2pi/(1-two*lambda_v1))
          ! the following weight corresponds to what is computed in the MC (i.e. standard RadISH implementation)
          ! might be slighlty inconsistent with the jetvheto (analytic) expansion
          ! wgt = Sudakov(L,res_order)*lumi_v1
          ! call interpolate(tt(:), lumi_nll(:),   histo_resummed(index)%values%right(i)/cs%Q, lumi_v1, dy_error, 3)
          ! wgt = wgt + Sudakov(L,res_order-1)*lumi_v1 * non_incl(cs%jet_radius,'all') * rp*two*cs%as2pi/(one-two*lambda_v1)
        end if

        !fill histogram for cumulant and its statistical error
        if (bin) then
          histo_resummed(index)%values%weight(i) = histo_resummed(index)%values%weight(i) + wgt*vegas_wgt*nev
          histo_resummed(index)%values%error(i)  = histo_resummed(index)%values%error(i)  + (wgt*vegas_wgt*nev)**2
        end if
      end do
    else
      do histo_index=1,number_of_histograms
        if (histo_resummed(histo_index)%values%scale /= scale_index) cycle
        if (histo_resummed(histo_index)%values%binvar /= binvar_index) cycle
        lumi_v1=zero ! These are luminosities evaluated at kt1 !!
        do i=1,size(histo_resummed(histo_index)%values%right(:))
          ! compute shower ingredients
          L=Ltilde_scalar(histo_resummed(histo_index)%values%right(i)/cs%Q,cs%p)
          rp  = Radp_v(L)
          lambda_v1=cs%alphas_muR*L*beta0

          if (res_order == order_NLL) then
            call interpolate(tt(:), lumi_nll(:),   histo_resummed(histo_index)%values%right(i)/cs%Q, lumi_v1, dy_error, 3)
            wgt = Sudakov(L,res_order)*lumi_v1
          else if (res_order == order_NNLL) then
            call interpolate(tt(:), lumi_nnll(:),   histo_resummed(histo_index)%values%right(i)/cs%Q, lumi_v1, dy_error, 3)
            ! the following weight is from the original JetVHeto implementation
            ! write(*,*) 'jetveto_resummation: using original JetVHeto implementation'
            wgt = Sudakov(L,res_order)*lumi_v1*(1 + &
            & non_incl(cs%jet_radius,'all') * rp*two*cs%as2pi/(1-two*lambda_v1))
            ! the following weight corresponds to what is computed in the MC (i.e. standard RadISH implementation)
            ! might be slighlty inconsistent with the jetvheto (analytic) expansion
            ! wgt = Sudakov(L,res_order)*lumi_v1
            ! call interpolate(tt(:), lumi_nll(:),   histo_resummed(histo_index)%values%right(i)/cs%Q, lumi_v1, dy_error, 3)
            ! wgt = wgt + Sudakov(L,res_order-1)*lumi_v1 * non_incl(cs%jet_radius,'all') * rp*two*cs%as2pi/(one-two*lambda_v1)
          end if

          !fill histogram for cumulant and its statistical error
          if (bin) then
            histo_resummed(histo_index)%values%weight(i) = histo_resummed(histo_index)%values%weight(i) + wgt*vegas_wgt*nev
            histo_resummed(histo_index)%values%error(i)  = histo_resummed(histo_index)%values%error(i)  + (wgt*vegas_wgt*nev)**2
          end if
        end do
      end do
    endif

  ! update the shower weight to be returned to VEGAS
  ! Pass the asymptotic contribution
  if( present(shower_wgt) ) then
    shower_wgt = shower_wgt + wgt
  endif

  end subroutine jetveto_resummation

end module shower_routines
